/*
  -------------------- file generate.c
  Created 18 Jun 94 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/* #define DEBUG */
#include "our_common.h"
#include "../libinclude/ntup.h"
#include "../general/coord.h"
#include "../general/pi.h"
#include "../general/teval.h"
#include "generate.h"
#include "indep.h"
#include "region.h"
#include "prog/matrix.h"
#include "prog/memalloc.h"
#include "prog/vector.h"

const struct coord_struct * s_debug_pi;

/*  Generate the initial set of regions for timeslicing.  The list
    is generated in a non-random order.  Currently this set is the set
    of all possible length 2 partial solutions.  We do not check that
    the partial solutions themselves satisfy the p and q constraint;
    they may not.

    Works even if one of the dimensions of pi has size 1.
*/
void
back2_gen_initial_regions(
const struct coord_struct *pi,
const int tuple_kind,   /* TUPLE_XX */
const int n,            /* interested in constructing n-tuples, n>1 */
/* we set *list_ret to region list, *nregions to number of regions */
struct region_struct **list_ret,
int *nregions_ret,
void  *(*allocfunc) P((const int))
)
{
  struct region_struct *regions;
  int nregions;
  int region_index;                 /* index into region list */
  const int pirows = pi->row_pos + 1;
  const int picols = pi->col_pos + 1;
  const int ncolors = 1 + (tuple_kind == TUPLE_BLACKWHITE);
  int start_color,end_color;

  if ((pirows < 2) && (picols < 2))
    fatal("get_initial_regions, rp or cp");
  if (n < 2)  fatal("get_initial_regions, n");

  /*  Compute the nubmer of regions.  Allocate the region array.
      Fill in the base_xx fields.
  */

  nregions = ncolors*ncolors*((pirows-1)*(picols-1) + pirows*picols-1);
  regions = (*allocfunc) (nregions*sizeof(struct region_struct));
  /*  Allocate space for fields of each region structure.
  */
  {
    int i;
    for (i = 0; i < nregions; i++)
      regions[i].ps = vec_alloc(n, allocfunc);
  }

  region_index = 0;
  if (tuple_kind == TUPLE_BLACKWHITE)
    {  start_color = 0;  end_color = 1;  }
  else if (tuple_kind == TUPLE_WHITEONLY)
    {  start_color = WHITE;  end_color = WHITE;  }
  else
    {  start_color = BLACK;  end_color = BLACK;  }

  /*  First kind of region. (Contains pixel in first row, pixel in
      first column, no pixel at origin.)
      Note: if pi is 1-dimensional, then there are no regions of this kind.
      Value of region_index set to 0 and saved for later.
  */
  {
    int acolor, bcolor;   /* colors of a and b pixels */
    int arow;             /* pixel a is at (arow,0) */
    int bcol;             /* pixel b is at (0,bcol) */

    for (arow = 1; arow < pirows; arow++) {
      for (bcol = 1; bcol < picols; bcol++) {
        for (acolor = start_color; acolor <= end_color; acolor++) {
          for (bcolor = start_color; bcolor <= end_color; bcolor++) {
            const int apiid                 = pi->rowcol_to_id[arow][0];
            const int apiidc                = PIID_C_TO_PIIDC(apiid,acolor);
            const int bpiid                 = pi->rowcol_to_id[0][bcol];
            const int bpiidc                = PIID_C_TO_PIIDC(bpiid,bcolor);
            struct region_struct *const reg = & regions[region_index++];

            reg->ps[0]                      = apiidc;
            reg->ps[1]                      = bpiidc;
            reg->id                         = region_index;
            reg->base_len                   = 2;
          }
        }
      }
    }
  }
  /* Second kind (contains pixel at origin)
     Value of region_index used from before.
  */
  {
    int acolor, bcolor;       /* colors of a and b pixels */
    const int apiid = pi->rowcol_to_id[0][0];   /* a is at (0,0) */
    int brow,bcol;            /* pixel b (not allowed to be at (0,0)) */

    for (brow = 0; brow < pirows; brow++) {
      for (bcol = 0; bcol < picols; bcol++) {
        for (acolor = start_color; acolor <= end_color; acolor++) {
          for (bcolor = start_color; bcolor <= end_color; bcolor++) {
            if ((brow != 0) || (bcol != 0)) {
              const int apiidc                = PIID_C_TO_PIIDC(apiid,acolor);
              const int bpiid                 = pi->rowcol_to_id[brow][bcol];
              const int bpiidc                = PIID_C_TO_PIIDC(bpiid,bcolor);
              struct region_struct *const reg = & regions[region_index++];
              reg->ps[0]                      = apiidc;
              reg->ps[1]                      = bpiidc;
              reg->id                         = region_index;
              reg->base_len                   = 2;
            }
          }
        }
      }
    }
  }

  ASSERT(nregions >= 1);

  /*  Set miscellaneous fields
  */
  {
    int i;
    for (i = 0; i < nregions; i++) {
      struct region_struct * const reg = & regions[i];

      reg->extend_len = 0;
      reg->nomore     = 0;
      teval_alloc(reg->teval, allocfunc);
    }
  }
  ASSERT(region_index == nregions);

  *list_ret = regions;
  *nregions_ret = nregions;
}

/*  Partial solution when origin is in first two components.
*/
static void
gen_candidates_origin(
const int tuple_kind,
const struct coord_struct * const pi,
const int non_origin_piid,      /* piid of non-origin pixel of first two */
int * const list_ret,
int * const list_len_ret
)
{
  const int pirows = pi->row_pos + 1;
  const int picols = pi->col_pos + 1;
  int norow, nocol;     /* row and column of the non-origin p.s. pixel */
  int i;                /* index into return list */
  int row, col;         /* row, column of current pixel */

  norow = pi->id_to_row[non_origin_piid];
  nocol = pi->id_to_col[non_origin_piid];
  ASSERT((norow > 0) || (nocol > 0));

  for (i = 0, row = norow; row < pirows; row++) {
    for (col = (row == norow) ? nocol+1 : 0; col < picols; col++) {
      const int piid = pi->rowcol_to_id[row][col];

      if ((tuple_kind==TUPLE_BLACKWHITE) || (tuple_kind==TUPLE_WHITEONLY))
        list_ret[i++] = PIID_C_TO_PIIDC(piid,WHITE);
      if ((tuple_kind==TUPLE_BLACKWHITE) || (tuple_kind==TUPLE_BLACKONLY))
        list_ret[i++] = PIID_C_TO_PIIDC(piid,BLACK);
    }
  }

  *list_len_ret = i;
  if (tuple_kind == TUPLE_BLACKWHITE)
    ASSERT(i == 2*((pirows-norow-1)*picols + (picols-nocol-1)) );
  else
    ASSERT(i == (pirows-norow-1)*picols + (picols-nocol-1));
}

/*  Partial solution when origin is not in first two components.
*/
static void
gen_candidates_no_origin(
const int tuple_kind,
const struct coord_struct * const pi,
const int * const ps,   /* first two piidc's */
int * const list_ret,
int * const list_len_ret
)
{
  const int pirows = pi->row_pos + 1;
  const int picols = pi->col_pos + 1;
  int leftrow;          /* row of pixel in column 0 */
  int topcol;           /* column of pixel in row 0 */
  int i;                /* index into return list */
  int row, col;         /* row, column of current pixel */

  {
    const int piid0 = PIIDC_TO_PIID(ps[0]);
    const int row0  = pi->id_to_row[piid0];
    const int col0  = pi->id_to_col[piid0];
    const int piid1 = PIIDC_TO_PIID(ps[1]);
    const int row1  = pi->id_to_row[piid1];
    const int col1  = pi->id_to_col[piid1];
    if (row0 == 0)          /* ps[0] is in first row */
      {  topcol = col0;  leftrow = row1;  ASSERT(col1 == 0);  }
    else                    /* ps[1] is in first row */
      {  topcol = col1;  leftrow = row0;  ASSERT((col0==0) && (row1==0));  }
  }

  for (i = 0, row = 0; row < pirows; row++) {
    for (col = 0; col < picols; col++) {
      if ((row == 0) && (col <= topcol))  continue;
      if ((col == 0) && (row <= leftrow))  continue;
      {
        const int piid = pi->rowcol_to_id[row][col];

        if ((tuple_kind==TUPLE_BLACKWHITE) || (tuple_kind==TUPLE_WHITEONLY))
          list_ret[i++] = PIID_C_TO_PIIDC(piid,WHITE);
        if ((tuple_kind==TUPLE_BLACKWHITE) || (tuple_kind==TUPLE_BLACKONLY))
          list_ret[i++] = PIID_C_TO_PIIDC(piid,BLACK);
      }
    }
  }

  *list_len_ret = i;
  if (tuple_kind == TUPLE_BLACKWHITE)
    ASSERT(i==2*((pirows-leftrow-1)+(picols-topcol-1)+(pirows-1)*(picols-1)));
  else
    ASSERT(i==((pirows-leftrow-1)+(picols-topcol-1)+(pirows-1)*(picols-1)));
}

/*  Generate the list of piidc's associated with a length 2
    partial solution.  These are the piidc's that can be used
    to draw the remainder of the solution from.
*/
void
back2_gen_region_piidc_list(
const int tuple_kind,
const struct coord_struct * const pi,
const int * const ps,             /* first two piidc's */
int * const list_ret,
int * const list_len_ret
)
{
  const int apiidc      = ps[0];
  const int apiid       = PIIDC_TO_PIID(apiidc);
  const int arow        = pi->id_to_row[apiid];
  const int acol        = pi->id_to_col[apiid];

  const int bpiidc      = ps[1];
  const int bpiid       = PIIDC_TO_PIID(bpiidc);
  const int brow        = pi->id_to_row[bpiid];
  const int bcol        = pi->id_to_col[bpiid];

#ifdef DEBUG
  s_debug_pi = pi;
#endif  /*DEBUG*/

  if ((arow == 0) && (acol == 0))
    gen_candidates_origin(tuple_kind,pi,bpiid,list_ret,list_len_ret);
  else if ((brow == 0) && (bcol == 0))
    gen_candidates_origin(tuple_kind,pi,apiid,list_ret,list_len_ret);
  else
    gen_candidates_no_origin(tuple_kind,pi,ps,list_ret,list_len_ret);
}

/*  qsort-style function for region_struct.
    Basically a front end for teval_sort.
*/
static int
region_compare(
const void * const p1,
const void * const p2
)
{
  const struct region_struct * const r1 = p1;
  const struct region_struct * const r2 = p2;

  return(teval_compare(r1->teval,r2->teval));
}

/*  Sort a list of regions according to their evaluation functions
    of their bases.  Because this sorting goes on every time p
    changes, a given partially searched region may switch between
    active and inactive arbirtarily many times.  (Its evaluation
    function depends on p, and it may move around in the region
    list when p changes.)

    We permute the regions of sorted list, somewhat randomly.

    Note: if a region fails, it is completely thrown away.  Thus
    allocated fields of the region structure are lost.

    Note: it is possible for this function to set *nregions to 0.
*/
void
back2_region_sort(
struct region_struct *const regions,
int *const nregions
)
{
  int i;
  int len = *nregions;

  /* Evaluate each region.  Delete regions that failed.
  */
  for (i = 0; i < len; ) {
    struct region_struct * const reg = & regions[i];
    const int passed = teval_from_scratch(reg->ps, reg->base_len, reg->teval);
    if (! passed) {
      regions[i] = regions[len-1];    /* structure assignment */
      len--;
    }
    else
      i++;
  }

  if (len > 1) {
    qsort(regions, len, sizeof(regions[0]), region_compare);
    bsbegin();
    {
      int * const perm                  = vec_alloc(len, bsalloc);
      struct region_struct * const buff = bsalloc(sizeof(buff[0]) * len);
      int i;
      long random();

      indep_perm(perm,len,random);
      for (i = 0; i < len; i++)
        buff[i] = regions[perm[i]];
      for (i = 0; i < len; i++)
        regions[i] = buff[i];
    }
    bsend();
  }

#ifdef DEBUG
  {
    int j;
    printf("------------------------- begin semi-sorted list in back2_region_sort\n");

    for (j = 0; j < len; j++) {
      const struct region_struct * const reg = & regions[j];
      printf("----------- region %d\n", j);
      printf("-- tuple\n");
      ps_print(stdout,s_debug_pi,reg->ps,reg->base_len,1);
      printf("-- evaluation\n");
      debug_teval(NULL, reg->teval);
    }
    printf("------------------------- end sorted list\n");
  }
#endif /*DEBUG*/

  *nregions = len;
}
