/*
  -------------------- file region.c
  Created 19 Jun 94 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Carry out the search for a region of the search space.
  Functions starting with setup_ are used to carry out the context
  switch to the current region of the search space.
  Note that level -1 is sometimes referred to.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <limits.h>

/* #define DEBUG */
/* #define TEMP */

#define UTEST_INTERVAL  99  /* check for update every UTEST_INTERVAL nodes */

#include "our_common.h"
#include "../libinclude/ntup.h"
#include "../general/pi.h"
#include "../general/tmask.h"
#include "back2int.h"
#include "generate.h"
#include "indep.h"
#include "region.h"
#include "prog/bitarray.h"
#include "prog/bitpar.h"
#include "prog/memalloc.h"
#include "prog/random.h"
#include "prog/running.h"
#include "prog/vector.h"

/*  Structure used to keep the information we need at each level.
*/
struct levinfo_struct {
  int num_candidates; /* length of candidate list */
  int *can_list;      /* candidates for lev i are at [i-1] */
                      /* levels > -1 are sorted; level -1 is not sorted */
  int index;          /* current index into previous level's candidate list */
  int piidc;          /* levinfo[i-1].can_list[lev[i].index] */
};

/*  Structure used for sorting candidates.
*/
struct cansort_struct {
  int piidc;
  struct teval_struct teval[1];
};

#ifdef HEAVY_ASSERTIONS_ON
int region_assertions_on = 1;
#else
int region_assertions_on = 0;
#endif /*HEAVY_ASSERTIONS_ON*/

/* Information that is set when search_region() is called, and remains
   valid until search_region() is exited.  The prefix "s_" is used
   to set these variables apart from other ones.  It stands for "static".
*/
static int s_n;
static int s_p;
static int s_q;
/*  Number of levels needed to hold bit parallel representation of "n".
    This is big enough to hold any needed q values.
*/
static const struct coord_struct *s_pi;
static int s_tuple_kind;
static FILE *s_f;
static int *s_base;             /* piidc's of base part */
static int s_base_len;
static int s_max_index;         /* calculated from width */
static int s_max_candidates;
static struct cansort_struct *s_cansort_buff;
static unsigned s_regionseed;
static unsigned s_ioflags;

/*  Indexed by level.
    [-1 .. (s_n-1 - base_len)]
*/
struct levinfo_struct *s_levinfo;

/*  Statistics.
*/
static double *s_lev_nodes;   /* number of nodes at each level */

/*  qsort-style function for cansort_struct.
    Basically a front end for teval_sort.
*/
static int
cansort_compare(
const void * const p1,
const void * const p2
)
{
  const struct cansort_struct * const c1 = p1;
  const struct cansort_struct * const c2 = p2;

  return(teval_compare(c1->teval,c2->teval));
}

static void
make_pixels(
const int len,
struct pix_struct * const pix_ret
)
{
  int i;

  for (i = 0; i < len; i++) {
    int piidc;
    if (i < s_base_len)
      piidc           = s_base[i];
    else {
      piidc = s_levinfo[i].piidc;
    }
    piidc_to_pix(s_pi,piidc,& pix_ret[i]);
  }
}

static void
print_partial_solution(
const int len
)
{
  bsbegin();
  {
    struct pix_struct *const pixels = bsalloc (len * sizeof(pixels[0]));
    make_pixels(len,pixels);
    ntup_print(s_f,pixels,len,1);       /* 1 = long format */
  }
  bsend();
}

/*  Seed the random number generator lrand48.  First compute a hash of
    the partial solution vector.  Then use this as the seed to lrand48.

    For the purpose of hashing the vector, consider the region seed to
    be appended to the current partial solution vector.  This makes runs
    give different results when a different seed S is used for the random
    number generator, because the region seeds change when S changes.
*/
static void
our_seed(
const int lev
)
{
  bsbegin();
  {
    unsigned * const ps = vec_u_alloc(lev+1, bsalloc);
    int i;
    for (i = 0; i < lev; i++) {
      const int piidc = s_levinfo[i].piidc;
      ASSERT(piidc >= 0); /* conversion not well defined if piidc<0 */
      ps[i] = piidc;      /* conversion from signed to unsigned */
    }
    ps[lev] = s_regionseed;
    {
      const unsigned seed = vec_u_hash(ps, lev+1);
      srand48(seed);
#ifdef TEMP
fprintf(stdout,"@a%d\n", seed);
fprintf(s_f,"@a%d\n", seed);
#endif /*TEMP*/
    }
  }
  bsend();
}

/*  Set s_levinfo[lev].can_list and s_levinfo[lev].num_candidates.
    (These are the candidates available at level lev+1.)
    Assumes that the level "lev" positive and negative information
    has been updated, according to placing a pixel (piid, color)
    at level "lev".  (We do not care explicitly about color, so it
    is not an argument to us.)

    The new candidate list is a subset of the previous candidate list.
    The new candidate list is put at level "lev", the old one comes
    from level "lev-1".

    Notes: we use the space at level lev+1.  For each candidate from the
    previous list, we simulate adding the candidate at level "lev+1".
    Assumes the update has been done for level "lev" already.
*/
static void
can_make_list(
const int lev,
const int piid
)
{
  const int prev_num_candidates = s_levinfo[lev-1].num_candidates;
  const int * const prev_list   = s_levinfo[lev-1].can_list;
  int * const cur_list          = s_levinfo[lev].can_list;
  int newlen                    = 0;
  int i;
#ifdef TEMP
  static int id = 0;
#endif  /*TEMP*/

  ASSERT(lev < s_n-1);

  for (i = 0; i < prev_num_candidates; i++) {
    const int can_piidc = prev_list[i];
    const int can_piid  = PIIDC_TO_PIID(can_piidc);
    if (can_piid != piid) {
      const int passed = do_tupdate(lev+1,can_piidc);
      if (passed) {
        do_teval(s_cansort_buff[newlen].teval);
        s_cansort_buff[newlen++].piidc = can_piidc;
      }
    }
  }

  /*  Sort candidates in cansort_buff, then copy piidc's in sorted order
      to final destination.
  */
  if (newlen > 1)
    qsort(s_cansort_buff,newlen,sizeof(s_cansort_buff[0]),cansort_compare);
  
  for (i = 0; i < newlen; i++)
    cur_list[i] = s_cansort_buff[i].piidc;

  if (newlen > 1) {
    ASSERT(newlen+1 < s_max_candidates);

#ifdef TEMP
    id++;
    fprintf(stdout,"@@@id%d\n", id);
    fprintf(s_f,"@@@id%d\n", id);
#endif  /*TEMP*/
    our_seed(lev);
    bsbegin();
    {
      int * const perm = vec_alloc(newlen, bsalloc);
      int * const buff = vec_alloc(newlen, bsalloc);
      int i;

      indep_perm(perm,newlen,lrand48);
      for (i = 0; i < newlen; i++)
        buff[i] = cur_list[perm[i]];
      for (i = 0; i < newlen; i++)
        cur_list[i] = buff[i];
    }
    bsend();
  }
  
#ifdef OLD_CAN_LIMIT
  if ((lev > 1) && (newlen > 2*s_n))
    newlen = 2*s_n;
#endif /*OLD_CAN_LIMIT*/

  s_levinfo[lev].num_candidates = newlen;
}

static struct levinfo_struct *
setup_levinfo(
void  *(*af) P((const int))   /* allocfunc */
)
{
  struct levinfo_struct * levinfo;
  int lev;

  levinfo = (*af) ((s_n+1) * sizeof(levinfo[0]));
  levinfo++;                    /* indexed -1 .. s_n-1 */
  for (lev = -1; lev < s_n; lev++) {
    levinfo[lev].can_list = vec_alloc(s_max_candidates, af);
  }
  return(levinfo);
}

/*  Update level information, corresponding to placing a pixel
    piidc at level lev.
    Returns:
      1 if the new partial solution meets the p and q constraints
      0 if not
*/
static int
lev_update(
const int lev,   /* level (0 ... n-1) */
const int piidc
)
{
  const int passed = do_tupdate(lev,piidc);
  if (! passed)
    return(0);

  if (lev < s_n-1) {
    can_make_list(lev,PIIDC_TO_PIID(piidc));
    /*  There must be at least enough candidates to supply one candidate
        for each of the remaining levels.  If there aren't, then this
        node is a failure.  This is a generalization of checking to see
        if the current node has 0 candidates.
    */
    if (lev+1+s_levinfo[lev].num_candidates < s_n)
      return(0);
  }

  return(1);
}

/*  Set up backtracking information about each level.
    See if the current partial solution is satisfactory.

    We return:
      -1            do no more work on partial solution
      lev>=0        start work in extended solution at level lev
    We also have some side effects.

    First, we work on the base part of the partial solution.
    Then work on the extended part (if any).
*/
static int
setup_levels(
const int * const ps,
const int extend_len
)
{
  /* Base part.
  */
  {
    int lev;
    for (lev = 0; lev < s_base_len; lev++) {
      const int piidc       = ps[lev];
      const int passed      = lev_update(lev, piidc);
      s_levinfo[lev].index  = 0;
      s_levinfo[lev].piidc  = piidc;
      if (! passed)
        return(-1);                   /* termination for this region */
    }
  }

  /*  Extended part (if any).
  */
  if (extend_len == 0) {              /* no extended solution */
    ASSERT(s_levinfo[-1].num_candidates > 0);
    s_levinfo[s_base_len].index = 0;
    s_levinfo[s_base_len].piidc = s_levinfo[s_base_len-1].can_list[0];
    return(s_base_len);
  }
  else {                              /* extended solution present */
    int lev;
    for (lev = s_base_len; lev < s_base_len + extend_len; lev++) {
      const int piidc       = ps[lev];
      const int found_index = vec_find_scalar( s_levinfo[lev-1].can_list,
                                    s_levinfo[lev-1].num_candidates, piidc);
      ASSERT(s_levinfo[lev-1].num_candidates > 0);
      ASSERT((found_index != -1) && (found_index <= s_max_index));
      s_levinfo[lev].index  = found_index;
      s_levinfo[lev].piidc  = piidc;
      if (! lev_update(lev,piidc))  /* update failed */
        return(lev);
    }
    return(s_base_len+extend_len-1);
  }
}

/*  Save extended part of partial solution:
      (1) convert it to ( pi index , color index ) format
      (2) save in region structure
    Leave base part of partial solution alone.
*/
static void
save_partial_solution(
const int len,
struct region_struct *const region
)
{
  int i;

  region->extend_len = len - s_base_len;
  for (i = s_base_len; i < len; i++)
    region->ps[i]   = s_levinfo[i].piidc;
}

static void
status_update(
const int lev,
const struct to_region_struct * const to_region,
const double nodes,
int *const update_count
)
{
  const int update_interval = to_region->update_interval;

  if (update_interval != UPDATE_NONE) {
    int j;
    int do_update = 0;

    if (update_interval == UPDATE_OFTEN)
      do_update = 1;
    else if (update_interval > 0) {
      if (*update_count == 0)
        do_update = update_test();
        *update_count = ((*update_count>=UTEST_INTERVAL) ? 0 : *update_count+1);
    }
    if (do_update) {
      if (s_ioflags & IOF_NO_TIME_DEPENDENT_OUTPUT)
        fprintf(s_f,"----- Update\n");
      else
        update_header_short(s_f);
      fprintf(s_f,"At search node %g for p=%d, q=%d.  %d active regions.\n",
        to_region->p_nodes + nodes, s_p,s_q, to_region->nactive_regions);
      fprintf(s_f,"can indices: ");
      for (j = 0; j <= lev; j++)
        fprintf(s_f,"%5d", s_levinfo[j].index);
      fprintf(s_f,"\n");
      fprintf(s_f,"can_lengths: ");
      for (j = -1; j < lev; j++)
        fprintf(s_f,"%5d", s_levinfo[j].num_candidates);
      fprintf(s_f,"\n");
      fprintf(s_f,"piidc list:  ");
      for (j = 0; j <= lev; j++)
        fprintf(s_f,"%5d", s_levinfo[j].piidc);
      fprintf(s_f,"\n");
      fprintf(s_f, "Tuple:\n");
      print_partial_solution(lev+1);
      update_start_timer();
    }
  }
}

/*  backgen_next from ../general/backgen.c, adapted for use here.
    Returns 1 if we generated the next partial solution in the region,
            0 if the region is exhausted.
*/
static int
our_backgen_next(
const int oldpassed,                /* did previous tuple pass constraints? */
int * const lev_ret                 /* search level */
/*  Before the first call, the caller should set:
  *lev_ret = 0
  s_levinfo[0].can_index = 0
*/
)
{
  int index;

  if (oldpassed && (*lev_ret < s_n-1))
    {  (*lev_ret)++;  index = 0;  goto good;  }

  do {
    const int newindex = s_levinfo[*lev_ret].index+1;
    if (newindex <= MIN(s_max_index,s_levinfo[(*lev_ret)-1].num_candidates-1))
      {  index = newindex;  goto good;  }
    (*lev_ret)--;
  } while ((*lev_ret) >= s_base_len);
  return(0);                      /* failure */

good:
  s_levinfo[*lev_ret].index = index;
  s_levinfo[*lev_ret].piidc = s_levinfo[*lev_ret-1].can_list[index];
  return(1);                      /* success */
}

static struct cansort_struct *
init_cansortbuff(
const int len,
void  *(*allocfunc) P((const int))    /* allocfunc */
)
{
  struct cansort_struct * const l = (*allocfunc) (sizeof(l[0]) * len);
  int i;

  for (i = 0; i < len; i++)
    teval_alloc(l[i].teval,allocfunc);
  return(l);
}

/*  Note: the values of p and q can change between invocations of
    this function, even for the same region in the same problem.
    So, one of the first things we do is test the extent to which the
    current partial solution satisfies the given p and q constraints.
    Note also that the region generator does not check whether the
    base part of the partial solution satisfies the constraints, so
    it may well not.  This is another reason for us doing the
    checking at the start of this function.

    If we find a solution, we do one of two things.
    (1)  If it is the first solution found for a given value of p,
         we return the solution to our caller (and we exit).
    (2)  Otherwise, we increment the solution count and continue.
*/
void
back2_search_region(
const struct to_region_struct * const to_region,
struct region_struct * const region,
struct from_region_struct * const from_region,
struct pix_struct * const solution_ret
)
{
  int update_count = 0;
  int lev;                  /* current search level */
  /* here "node" refers to a node in the extended part of the p.s. */
  const double max_slice_nodes = to_region->max_slice_nodes;
  double nodes;             /* new nodes encountered during this timeslice */
  double solutions;         /* how many solutions we found */

  bsbegin();

  /* Copy commonly used information into global (static to this file) vars.
     More s_ variables are set up a little farther down below.
  */
  s_n                 = to_region->n;
  s_p                 = to_region->p;
  s_q                 = to_region->q;
  s_pi                = to_region->pi;
  s_lev_nodes         = to_region->lev_nodes;
  s_tuple_kind        = to_region->tuple_kind;
  s_f                 = to_region->f;
  s_ioflags           = to_region->ioflags;
  s_base              = vec_alloc(s_n, bsalloc);
  s_base_len          = region->base_len;
  vec_copy(s_base, & region->ps[0], s_base_len);
  s_max_index         = (to_region->width ? to_region->width-1 : INT_MAX-1);
  s_max_candidates    = can_determine_max(s_pi)+1;
    /* the +1 is because can_make_list() uses an extra slot for the seed. */
  s_regionseed        = region->seed;
  s_cansort_buff      = init_cansortbuff(s_max_candidates,bsalloc);

  nodes = 0;
  solutions = 0;
  if (region->nomore)           /* very rare special case */
    {  from_region->status = SEARCH_EXHAUSTED;  goto done;  }

  /*  Allocate and initialize level information
  */
  s_levinfo = setup_levinfo(bsalloc);
  back2_gen_region_piidc_list(s_tuple_kind,s_pi,region->ps,
              s_levinfo[-1].can_list, & s_levinfo[-1].num_candidates);
  if (s_levinfo[-1].num_candidates == 0)    /* search region is empty */
    {  from_region->status = SEARCH_EXHAUSTED;  goto done;  }

  /*  Set up backtracking information, using base and extended part of
      current solution.
  */
  lev = setup_levels(region->ps,region->extend_len);
  if (lev == -1)
    {  from_region->status = SEARCH_EXHAUSTED;  goto done;  }

  /*  The first time through the loop, we will do one of three things.
      1.  If the partial solution was length 0, then we will start
          from scratch.  (This may not ever happen.)
      2.  If it was found above that the current p.s. fails, then we
          start the search at at the level that caused the failure.
      3.  If the current p.s. succeeds, then we will start at the
          deepest level of the partial solution.  If there was an
          extended part, then the deepest level (level lev) was
          checked in setup_levels.  It is also checked below.
          This is unnecesary, but does not hurt anything.
          (Care is needed in the nodes calculation.)
  */
  while (1) {
    int passed; /* whether or not the current partial solution passed */

    ASSERT((lev >= s_base_len) && (lev < s_n));
    ASSERT(s_levinfo[lev].index >= 0);
    ASSERT(s_levinfo[lev].index < s_levinfo[lev-1].num_candidates);
    ASSERT(s_levinfo[lev].index <= s_max_index);

    nodes++;
    s_lev_nodes[lev]++;

    /* periodic status updates */
    status_update(lev,to_region,nodes,&update_count);

    passed = lev_update(lev,s_levinfo[lev].piidc);
    if (passed && (lev+1 == s_n)) {
      const int print = to_region->back_print;

      solutions++;
      if (    ((print == BACK_PRINT_FIRST) && (! to_region->found_first)) ||
              (print == BACK_PRINT_ALL) || (print == BACK_PRINT_CANON_ONLY) ) {
        fprintf(s_f, "--- Solution for p=%d, q=%d:\n", s_p, s_q);
        print_partial_solution(lev+1);
      }
      if (! to_region->found_first) {
        make_pixels(s_n,solution_ret);
        /*  Prepare to leave.  Move on to the next partial solution,
            then save the state of the search.
        */
        if (our_backgen_next(passed,&lev))
          save_partial_solution(lev+1,region);
        else
          region->nomore = 1;  /* next time in this region, quit immediately */
        from_region->status = SEARCH_FOUND_FIRST;
        goto done;
      }
    }
    /*  Get next partial solution, if any.
    */
    if (! our_backgen_next(passed,&lev))
      {  from_region->status = SEARCH_EXHAUSTED;  goto done;  }

    if ((max_slice_nodes > 0) && (nodes >= max_slice_nodes)) {
      /*  Suspend the search in this region of the search space.  The
          current partial solution has not yet been tested.  It will
          be tested the first time through the loop when the search
          resumes in this region.
      */
      save_partial_solution(lev+1,region);
      from_region->status = SEARCH_INCONCLUSIVE;
      goto done;
    }
  }

done:
  from_region->nodes      = nodes;
  from_region->solutions  = solutions;

  bsend();
}
