/*
  -------------------- file charutil.c
  Created 19 May 94 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Utility functions for manipulating characters.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <limits.h>

#include "our_common.h"
#include "charutil.h"
#include "prog/io.h"
#include "prog/memalloc.h"
#include "prog/matrix.h"
#include "../libinclude/ntup.h"

/*  Front end to mat_uchar_trim2 for character_struct structures.
    Trim a character.  Allocate a new character that is the trimmed
    character, return it.  Declaration is in ntup.h.
*/
struct character_struct *
ntup_char_trim(
struct character_struct * const oldchar,  /* really is const * */
const int bgcolor,
void  *(*allocfunc) P((const int))
)
{
  struct character_struct *newchar;

  newchar = (*allocfunc) (sizeof(struct character_struct));
  mat_uchar_trim2(
    oldchar->a,oldchar->nrows,oldchar->ncols, bgcolor,
    & newchar->a, & newchar->nrows, & newchar->ncols, allocfunc);
  return(newchar);
}

/*  Print a graphical representation of a character's pixels.
*/
void
ntup_print_char_pixels(
FILE * const f,
const struct character_struct * const h
)
{
  int row,col;

  for (row = 0; row < h->nrows; row++) {
    for (col = 0; col < h->ncols; col++) {
      const int v = h->a[row][col];
      if (col > 0)
        fprintf(f, " ");
      if (v == WHITE)
        fprintf(f, ".");
      else if (v == BLACK)
        fprintf(f, "X");
      else if (v == PSEUDOCOLOR_NOMATCH)
        fprintf(f, "-");
      else if (v == PSEUDOCOLOR_MATCH)
        fprintf(f, "+");
    }
    fprintf(f,"\n");
  }
}

/*
    Returns: 0 if translations do not form a rectangular region.
             1 if they do.  In this case *.._ret are set.
    
    This function is available to other parts of the library, but not
    to the application.
*/
int
trans_form_rectangle(
const struct character_struct * const h,
int * const mindr_ret,
int * const maxdr_ret,
int * const mindc_ret,
int * const maxdc_ret
)
{
  const int ntrans = h->ntrans;
  int i;
  int mindr = INT_MAX;
  int mindc = INT_MAX;
  int maxdr = INT_MIN;
  int maxdc = INT_MIN;
  int done_min;

  for (i = 0; i < ntrans; i++) {
    const int dr = h->trans_list[i].dr;
    const int dc = h->trans_list[i].dc;
    if (dr < mindr)  mindr = dr;
    if (dr > maxdr)  maxdr = dr;
    if (dc < mindc)  mindc = dc;
    if (dc > maxdc)  maxdc = dc;
  }

  bsbegin();
  {
    const int rows = maxdr-mindr+1;
    const int cols = maxdc-mindc+1;
    unsigned char ** m = mat_uchar_alloc2(rows,cols,bsalloc);

    mat_uchar_zero2(m,rows,cols);
    for (i = 0; i < ntrans; i++) {
      const int dr = h->trans_list[i].dr;
      const int dc = h->trans_list[i].dc;
      m[dr-mindr][dc-mindc] = 1;
    }
    mat_uchar_minmax2(m,rows,cols,&done_min,NULL);

  }
  bsend();

  if (done_min == 0)
    return(0);

  if (mindr_ret) *mindr_ret = mindr;
  if (maxdr_ret) *maxdr_ret = maxdr;
  if (mindc_ret) *mindc_ret = mindc;
  if (maxdc_ret) *maxdc_ret = maxdc;

  return(1);
}

/*  Print the translations of a character.  If the translations form a
    rectangular region, print a compact representation of the region.
*/
void
print_char_trans(
FILE * const f,
const struct character_struct * const h
)
{
  int i;
  int mindr,maxdr,mindc,maxdc;
  const int ntrans = h->ntrans;

  if (ntrans == 0)
    {  fprintf(f, "(none)\n");  return;  }

  if (trans_form_rectangle(h,&mindr,&maxdr,&mindc,&maxdc))
    fprintf(f, "(%d .. %d) x (%d .. %d)\n", mindr, maxdr, mindc, maxdc);
  else {                  /* list representation */
    for (mindr=mindc=INT_MAX, maxdr=maxdc=INT_MIN, i = 0; i < ntrans; i++) {
      const int dr = h->trans_list[i].dr;
      const int dc = h->trans_list[i].dc;
      if (dr < mindr)  mindr = dr;
      if (dr > maxdr)  maxdr = dr;
      if (dc < mindc)  mindc = dc;
      if (dc > maxdc)  maxdc = dc;
    }
    {
      const int rwidth = MAX(fieldlen(mindr),fieldlen(maxdr));
      const int cwidth = MAX(fieldlen(mindc),fieldlen(maxdc));
      const int width = rwidth+cwidth+3;
      int outcol;

      for (i = 0, outcol = 0; i < ntrans; i++) {
        const int dr = h->trans_list[i].dr;
        const int dc = h->trans_list[i].dc;
        if (outcol+width >= OUTPUT_COLUMNS)
          {  fprintf(f, "\n");  outcol = 0;  }
        if (outcol > 0)  {  fprintf(f, " ");  outcol += 1;  }
        fprintf(f, "(%*d,%*d)", rwidth,dr,cwidth,dc);
        outcol += width;
      }
      fprintf(f, "\n");
    }
  }
}
