/*
  -------------------- file pi.c
  Created 17 May 94 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Functions concerning pi - the region from which pixels are selected.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#include <limits.h>

#define COORD_DEBUG
#include "our_common.h"
#include "../libinclude/ntup.h"
#include "coord.h"
#include "pi.h"
#include "prog/matrix.h"
#include "prog/memalloc.h"
#include "prog/permute.h"
#include "prog/vector.h"

#ifdef COORD_DEBUG
void
debug_coord(
const struct coord_struct *const coord
)
{
  const int nrows = coord->row_pos - coord->row_neg + 1;
  const int ncols = coord->col_pos - coord->col_neg + 1;

  printf("-----   coord\n");
  printf("%d ids   rows are %d .. %d   cols are %d .. %d\n",
    coord->nids,coord->row_neg,coord->row_pos,coord->col_neg,coord->col_pos);
  printf("rowcol_to_id:\n");
  mat_print2(stdout,coord->rowcol_to_id,nrows,ncols);
  printf("id_to_row:\n");
  vec_print(stdout,coord->id_to_row,coord->nids);
  printf("id_to_col:\n");
  vec_print(stdout,coord->id_to_col,coord->nids);
}

void
debug_piidc_vec(
const struct coord_struct * const pi,
const char * const str,
const int * const piidcs,
const int len
)
{
  int i;

  printf("-- begin piidc vec '%s'   len=%d:\n", str ? str : "", len);
  printf("   num   row col  color  piid  piidc\n");
  for (i = 0; i < len; i++) {
    const int piidc = piidcs[i];
    const int piid  = PIIDC_TO_PIID(piidc);
    const int row = pi->id_to_row[piid];
    const int col = pi->id_to_col[piid];
    const int color = PIIDC_TO_COLOR(piidc);
    printf("%6d%6d%4d%7d%6d%7d\n", i, row, col, color, piid, piidc);
  }
  printf("-- end piidc vec\n");
}
#endif /*COORD_DEBUG*/

/*  Fill in the fields of a pi structure, using the standard
    (non-random) row major square numbering.
*/
void
pi_make_standard(
const int rows,
const int cols,
struct coord_struct * const pi_ret,       /* we fill in fields of this */
void  *(*allocfunc) P((const int))
)
{
  /* local information that we move later */
  int nids;
  int **rowcol_to_id;
  int *id_to_row;
  int *id_to_col;

  nids = rows * cols;

  /*  Store conversions of pi coordinates: (square) <---> (row,col).
      Use row major ids.
  */
  {
    int row,col,id;

    rowcol_to_id = mat_alloc2(rows,cols,allocfunc);
    id_to_row = vec_alloc(nids,allocfunc);
    id_to_col = vec_alloc(nids,allocfunc);

    id = 0;
    for (row = 0; row < rows; row++) {
      for (col = 0; col < cols; col++) {
        rowcol_to_id[row][col] = id;
        id_to_row[id] = row;
        id_to_col[id] = col;
        id++;
      }
    }
    ASSERT(id == nids);
  }

  pi_ret->row_neg = 0;
  pi_ret->row_pos = rows-1;
  pi_ret->col_neg = 0;
  pi_ret->col_pos = cols-1;
  pi_ret->nids = nids;
  pi_ret->rowcol_to_id = rowcol_to_id;
  pi_ret->id_to_row = id_to_row;
  pi_ret->id_to_col = id_to_col;

  D(printf("standard pi:\n");)
  D(debug_coord(pi_ret);)
}

/*  Given one numbering of pi, form another one by randomly permuting
    the squares.  Fills in the fields of the new pi structure.

    We do not use bsalloc, so it is allowable for a caller
    to specify it as allocfunc.
*/
void
pi_make_random(
struct coord_struct *dest,
const struct coord_struct *src,
void  *(*allocfunc) P((const int))
)
{
  const int nids = src->nids;
  const int nrows = (src->row_pos - src->row_neg) + 1;
  const int ncols = (src->col_pos - src->col_neg) + 1;
  int *perm;    /* random permutation of 0 .. nids-1 */
  int **rowcol_to_id;
  int *id_to_row, *id_to_col;
  int oldid;
  long random();

  perm = vec_alloc(nids, xmalloc);
  randperm(perm, nids, random);
  rowcol_to_id = mat_alloc2(nrows,ncols,allocfunc);
  id_to_row = vec_alloc(nids,allocfunc);
  id_to_col = vec_alloc(nids,allocfunc);

  for (oldid = 0; oldid < nids; oldid++) {
    const int newid = perm[oldid];
    const int row = src->id_to_row[oldid];
    const int col = src->id_to_col[oldid];
    rowcol_to_id[row][col] = newid;
    id_to_row[newid] = row;
    id_to_col[newid] = col;
  }

  dest->row_neg = src->row_neg;
  dest->row_pos = src->row_pos;
  dest->col_neg = src->col_neg;
  dest->col_pos = src->col_pos;
  dest->nids = src->nids;
  dest->rowcol_to_id = rowcol_to_id;
  dest->id_to_row = id_to_row;
  dest->id_to_col = id_to_col;

  vec_free(perm,xfree);
}

void
piidc_to_pix(
const struct coord_struct * const pi,
const int piidc,
struct pix_struct * const pix_ret
)
{
  const int piid  = PIIDC_TO_PIID(piidc);
  const int color = PIIDC_TO_COLOR(piidc);
  const int row   = pi->id_to_row[piid];
  const int col   = pi->id_to_col[piid];

  pix_ret->row    = row;
  pix_ret->col    = col;
  pix_ret->color  = color;
}

/*  Convert from vector of piidc's to pixel format
*/
void
convert_ps_to_pixels(
const struct coord_struct *const pi,
const int * const ps,     /* components are piidc's */
const int ps_len,
struct pix_struct * const pix_ret
)
{
  int i;

  for (i = 0; i < ps_len; i++) {
    const int piidc = ps[i];
    piidc_to_pix(pi, piidc, & pix_ret[i]);
  }
}

/*  Print a vector of piidc's in ntup_print() format.
*/
void
ps_print(
FILE *f,
const struct coord_struct *const pi,
const int * const ps,   /* components are piidc's */
const int ps_len,       /* length of ps vector */
const int format        /* format to pass to ntup_print */
)
{
  bsbegin();
  {
    struct pix_struct *const pixels = bsalloc (ps_len * sizeof(pixels[0]));
    convert_ps_to_pixels(pi,ps,ps_len,pixels);
    ntup_print(f,pixels,ps_len,format);
  }
  bsend();
}
