/*
  -------------------- file tmask.c
  Created 17 May 94 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1994 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  "tmask" refers to translation mask.  For each character, for
  each color, for each square of pi, there is a "translation mask".
  There is a bit in the mask for each translation.
  The bit is 1 when a given pi square of the given color is the same
  color as the character, when the given pi square is translated
  by the translation corresponding to the bit.

  USERS MUST CALL bitword_global_init() before calling these routines.
*/

/* #define DEBUG */

#include "our_common.h"
#include "coord.h"
#include "tmask.h"
#include "../libinclude/ntup.h"
#include "prog/bitarray.h"
#include "prog/bitpar.h"
#include "prog/matrix.h"
#include "prog/memalloc.h"
#include "prog/vector.h"

#ifdef DEBUG
#include "pi.h"   /* debug_pi() */
void
debug_tmask(
const struct tmask_struct *tmask
)
{
  const int npi_ids = tmask->pi->nids;
  const int nchars = tmask->nchars;
  const int *ntrans = tmask->ntrans;
  const int *ntrans_words = tmask->ntrans_words;
  BITWORD ****mask;
  int piid,ch,color;
  int w;

  mask = tmask->mask;

  printf("---------------------- begin tmask\n");
  printf("maxtrans=%d  nchars=%d  ntrans[] = ", tmask->maxtrans,tmask->nchars);
  vec_print(stdout, ntrans, nchars);
  printf("ntrans_words[] = ");
  vec_print(stdout, ntrans_words, nchars);
  printf("pi:\n");
  debug_coord(tmask->pi);
  for (ch = 0; ch < nchars; ch++) {
    printf("---------- character %d\n", ch);
    for (piid = 0; piid < npi_ids; piid++) {
      printf("  -- pi id %d (at %d,%d)\n",
        piid,tmask->pi->id_to_row[piid],tmask->pi->id_to_col[piid]);
      for (color = 0; color < 2; color++) {
        if (color == 0)   printf("    --- white\n");
        else              printf("    --- black\n");
        for (w = 0; w < ntrans_words[ch]; w++) {
          printf("           ");
          bit_print_binary(stdout,& mask[piid][color][ch][w], BITS_PER_WORD);
          printf("\n");
        }
        printf("\n");
      }
    }
  }
  printf("---------------------- end tmask\n");
}

#endif  /*DEBUG*/

/*  Allocate space for a new coord_struct structure and its fields,
    and copy the data in a given coord_struct structure to the new structure.
*/
static struct coord_struct *
coord_clone(
const struct coord_struct *oc,
void  *(*allocfunc) P((const int))
)
{
  const int rows = oc->row_pos - oc->row_neg + 1;
  const int cols = oc->col_pos - oc->col_neg + 1;
  struct coord_struct *nc;

  nc = (*allocfunc) (sizeof(struct coord_struct));
  nc->row_neg       = oc->row_neg;
  nc->row_pos       = oc->row_pos;
  nc->col_neg       = oc->col_neg;
  nc->col_pos       = oc->col_pos;
  nc->nids          = oc->nids;
  nc->rowcol_to_id  = mat_alloc2(rows,cols,allocfunc);
    mat_copy2(oc->rowcol_to_id,rows,cols,nc->rowcol_to_id);
  nc->id_to_row     = vec_alloc(nc->nids, allocfunc);
    vec_copy(nc->id_to_row,oc->id_to_row,nc->nids);
  nc->id_to_col     = vec_alloc(nc->nids, allocfunc);
    vec_copy(nc->id_to_col,oc->id_to_col,nc->nids);
  return(nc);
}

/*
  Make the masks for a given set of characters and a given pi region.
  Note: no way to free explicitly (using allocfunc=bsalloc is recommended).
*/
void
tmask_make(
const struct character_struct *const chars,
const int nchars,
const struct coord_struct *const pi,
const int background,                       /* white,black,no-match,match */
struct tmask_struct *const tmask_ret,       /* we set fields x : tmask_ret->x*/
void  *(*allocfunc) P((const int))
)
{
  const int npiids = pi->nids;
  int * const ntrans = vec_alloc(nchars, allocfunc);
  int * const ntrans_words = vec_alloc(nchars, allocfunc);
  int piid;
  int ch;
  BITWORD ****mask;

  /*  Fill in ntrans and ntrans_words
  */
  for (ch = 0; ch < nchars; ch++) {
    ntrans_words[ch] = BIT_WORDS(chars[ch].ntrans);
    ntrans[ch] = chars[ch].ntrans;
  }

  /* Allocate mask array
  */
  mask = (*allocfunc) (npiids * sizeof(mask[0]));
  for (piid = 0; piid < npiids; piid++) {
    mask[piid] = (*allocfunc) (2 * sizeof(mask[0][0]));
    mask[piid][0] = (*allocfunc) (nchars * sizeof(mask[0][0][0]));
    mask[piid][1] = (*allocfunc) (nchars * sizeof(mask[0][0][0]));
    for (ch = 0; ch < nchars; ch++) {
      mask[piid][0][ch] = bit_alloc(ntrans[ch], allocfunc);
      mask[piid][1][ch] = bit_alloc(ntrans[ch], allocfunc);
    }
  }

  /*  Initialization
  */
  for (piid = 0; piid < npiids; piid++) {
    const int pi_row = pi->id_to_row[piid];
    const int pi_col = pi->id_to_col[piid];
    for (ch = 0; ch < nchars; ch++) {
      const int ch_nrows = chars[ch].nrows;
      const int ch_ncols = chars[ch].ncols;
      unsigned char **a;
      BITWORD *whitemask, *blackmask;
      int d;            /* translation number */

      a = chars[ch].a;
      whitemask = mask[piid][WHITE][ch];
      blackmask = mask[piid][BLACK][ch];
      bit_zero(whitemask,ntrans[ch]); /* zero unused bits (others set below) */
      bit_zero(blackmask,ntrans[ch]); /* zero unused bits (others set below) */
      for (d = 0; d < chars[ch].ntrans; d++) {
        const int drow = chars[ch].trans_list[d].dr;
        const int dcol = chars[ch].trans_list[d].dc;
        const int row = pi_row + drow;
        const int col = pi_col + dcol;
        int pixel;

        if ((row<0) || (row>=ch_nrows) || (col<0) || (col>=ch_ncols))
          pixel = background;
        else
          pixel = a[row][col];
        if (pixel == WHITE)
          {  BIT_SET(whitemask,d);  BIT_RESET(blackmask,d);  }
        else if (pixel == BLACK)
          {  BIT_SET(blackmask,d);  BIT_RESET(whitemask,d);  }
        else if (pixel == PSEUDOCOLOR_NOMATCH)
          {  BIT_RESET(whitemask,d);  BIT_RESET(blackmask,d);  }
        else if (pixel == PSEUDOCOLOR_MATCH)
          {  BIT_SET(whitemask,d);  BIT_SET(blackmask,d);  }
        else
          fatal("tmask_make");
      }   /* d */
    }     /* ch */
  }       /* piid */

  tmask_ret->nchars = nchars;
  tmask_ret->ntrans = ntrans;
  tmask_ret->ntrans_words = ntrans_words;
  tmask_ret->pi = coord_clone(pi,allocfunc);
  tmask_ret->mask = mask;
}

/*  Determine whether the positive (p) constraint is satisfied by
    a given n-tuple.
    Returns 0 if not satisfied, 1 if satisfied.
    Not written for blazing speed, but should be reasonably fast.
*/
int
tmask_pos_satisfied(
const struct pix_struct * const pixels,     /* row,column,colors of pixels */
const int n,                                /* number of pixels */
const int p,                                /* p constraint */
const struct tmask_struct * const pos_tmask /* precomputed pos masks */
)
{
  const int npos = pos_tmask->nchars; /* number of positive chars */
  int ch;           /* character we are interested in */
  BITWORD *curmask; /* and of masks for each pixel */
  BITWORD ****mask; /* pos_tmask->mask */
  int *pix_piid;    /* pi id's of tuple's pixels, for speed and convenience */
  int *pix_color;   /* colors of tuple's pixels, for speed and convenience */
  int maxtrans;

  bsbegin();
  D(printf("top of tmask_pos_satisfied\n");)

  /* note: we later make sure that the unused bits of curmask are 0,
     when we call bit_ones()
  */
  vec_minmax(pos_tmask->ntrans,pos_tmask->nchars,NULL,& maxtrans);
  curmask = bit_alloc(maxtrans, bsalloc);
  pix_color = vec_alloc(n,bsalloc);
  pix_piid = vec_alloc(n,bsalloc);
  mask = pos_tmask->mask;
  {
    /* make a table of pi id's and colors for fast, convenient reference */
    int **rowcol_to_piid;
    int pix;

    rowcol_to_piid = pos_tmask->pi->rowcol_to_id;
    for (pix = 0; pix < n; pix++) {
      const struct pix_struct *pp = & pixels[pix];
      pix_color[pix] = pp->color;
      pix_piid[pix] = rowcol_to_piid[pp->row][pp->col];
    }
  }
#ifdef DEBUG
  printf("ids:    ");
  vec_print(stdout,pix_piid,n);
  printf("colors: ");
  vec_print(stdout,pix_color,n);
#endif /*DEBUG*/
  for (ch = 0; ch < npos; ch++) {
    const int ntrans = pos_tmask->ntrans[ch];
    int pix;
    D(printf("character %d   ntrans=%d\n", ch, ntrans);)
    bit_ones(curmask,ntrans);
    for (pix = 0; pix < n; pix++) {
      D(printf("pix=%d  ntrans=%d  id=%d  color=%d\n",
        pix, ntrans, pix_piid[pix], pix_color[pix]);)
      bit_and(curmask,ntrans,curmask,mask[pix_piid[pix]][pix_color[pix]][ch]);
      /* COULDDO: check count here, in-line expand, maybe more */
#ifdef DEBUG
      printf("pixel %d.\n", pix);
      printf("  mask[id=%d][color=%d][ch=%d]=", pix_piid[pix],pix_color[pix],ch);
      bit_print_binary(stdout,mask[pix_piid[pix]][pix_color[pix]][ch],ntrans);
      printf("\n");
      printf("  after, curmask=");
      bit_print_binary(stdout,curmask,ntrans);
      printf("\n");
#endif /*DEBUG*/
    }
    D(printf("final count for ch %d is count=%d\n",
      ch, bit_count(curmask,ntrans));)
    if (bit_count(curmask,ntrans) < p)
      {  bsend();  return(0);  }
  }
  bsend();
  return(1);
}

/*  Determine whether the negative (q) constraint is satisfied by
    a given n-tuple.
    Returns 0 if not satisfied, 1 if satisfied.
    Not written for blazing speed, but should be reasonably fast.
    If q >= n then does virtually nothing.
*/
int
tmask_neg_satisfied(
const struct pix_struct * const pixels,     /* row,column,colors of pixels */
const int n,                                /* number of pixels */
const int q,                                /* q constraint */
const struct tmask_struct * const neg_tmask /* precomputed neg masks */
)
{
  if (q >= n)  return(1);
  {
    const int nneg = neg_tmask->nchars; /* number of negative chars */
    int ch;           /* character we are interested in */
    BITWORD *qpar;    /* bit-parallel representation of the number q */
    BITWORD *cursum;  /* running sum over current word of current pixel */
    const int levels = bitpar_levels_needed(n); /* about lg n */
    BITWORD ****mask; /* pos_tmask->mask */
    int *pix_piid;    /* pi id's of tuple's pixels, for speed, convenience */
    int *pix_color;   /* colors of tuple's pixels, for speed, convenience */

    /* at this point q < n; we compute number of levels based on n */
    bsbegin();

    qpar = bitpar_alloc(levels,bsalloc);
    cursum = bitpar_alloc(levels,bsalloc);
    pix_color = vec_alloc(n,bsalloc);
    pix_piid = vec_alloc(n,bsalloc);

    unsigned_to_bitpar(q,levels,qpar);  /* convert q to bit parallel repres. */
    mask = neg_tmask->mask;

    {
      /* make a table of pi id's and colors for fast, convenient reference */
      int **rowcol_to_piid;
      int pix;

      rowcol_to_piid = neg_tmask->pi->rowcol_to_id;
      for (pix = 0; pix < n; pix++) {
        const struct pix_struct *pp = & pixels[pix];
        pix_color[pix] = pp->color;
        pix_piid[pix] = rowcol_to_piid[pp->row][pp->col];
      }
    }

    for (ch = 0; ch < nneg; ch++) {
      const int nwords = neg_tmask->ntrans_words[ch];
      int pix,w;
      for (w = 0; w < nwords; w++) {
D(printf("----------ch=%d  word %d of %d\n",ch, w, nwords);)
        bitpar_zero(cursum,levels);
        for (pix = 0; pix < n; pix++) {
#ifdef DEBUG
          printf("-- pix=%d  before bitpar_add_onerow.\n", pix);
          printf("single:    ");
          bit_print_binary(stdout,& mask[pix_piid[pix]][pix_color[pix]][ch][w],32);
          printf("\n");
          {
            int i;
            for (i = 0; i < levels; i++) {
              printf("cursum[%d]: ",i);
              bit_print_binary(stdout,& cursum[i],32);
              printf("\n");
            }
          }
#endif  /*DEBUG*/
          bitpar_add_onerow(mask[pix_piid[pix]][pix_color[pix]][ch][w],cursum);
#ifdef DEBUG
          printf("-- after bitpar_add_onerow.\n");
          {
            int i;
            for (i = 0; i < levels; i++) {
              printf("cursum[%d]: ",i);
              bit_print_binary(stdout,& cursum[i],32);
              printf("\n");
            }
          }
          if (bitpar_compare_max(cursum, qpar, levels) > 0)   /* cursum > q */
            printf("failed - returning 0.\n");
#endif  /*DEBUG*/
        }
        /* Note: in the bitarray representation, unused bits are 0.
           The test below still works in this case.
        */
        if (bitpar_compare_max(cursum, qpar, levels) > 0)   /* cursum > q */
          {  bsend();  return(0);  }
      }
    }
    bsend();
    return(1);
  }
}
