/*
  -------------------- file ntup.h
  Created 16 Nov 93 by Andrew Shapira, Rensselaer Polytechnic Institute.
  Structures and defines for interface to/from the tuple finder.

	Copyright (C) 1993 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

#ifndef _181193_002646
#define _181193_002646

#define NTUPLIB_VERSION     "1.0"

/*  Information about a character.
*/
struct character_struct {
  int  nrows,ncols;
  unsigned char **a;
  int ntrans;     /* number of translations in list */
  struct {
    int dr, dc;   /* row, column movement */
  } *trans_list;
};

/*  Information about a pixel that is part of a tuple.
*/
struct pix_struct {
  int row, col;
  int color;
};

/*  Information about an n-tuple
*/
struct tuple_struct {
  int n;                      /* tuple is an n-tuple */
  int p,q;                    /* p,q value that n-tuple is designed for */
  struct pix_struct *pixels;  /* [i]: info about pixel i, i=0 .. n-1 */
};

/*  Structure for sending information to find_ntup().

    ----------------------------------------------------------------
    ----------------------------------------------------------------
    If you change this structure or associated constants,
    update report_input() and ntup_check_input().
    ----------------------------------------------------------------
    ----------------------------------------------------------------
*/
struct to_ntup_struct {
  int       n;                      /* look for tuples with n pixels */
  int       tuple_kind;             /* TUPLE_XX constant (black,white,bw) */
  int       pi_rows, pi_cols;       /* dims. of area tuples are drawn from */
  int       pq_kind;                /* PQ_XX constant */
  int       p,q;                    /* p,q to use if pq_kind==PQ_SPECIFIC */
  struct charclass_struct {
    int nchars;                     /* number of characters */
    struct character_struct *chars; /* list of characters */
    int bg;                         /* character background (a pseudocolor) */
  } pos, neg;                       /* positive and negative char. classes */
  int       update_interval;        /* UPDATE_XX, or positive num. of secs. */
  unsigned  ioflags;                /* mask with IOF_XX */
  FILE      *f;                     /* where to send generator-printed output */
};

/*  Structure for receiving information from find_ntup().

    ----------------------------------------------------------------
    ----------------------------------------------------------------
    If you change this structure or associated constants,
    update report_output().
    ----------------------------------------------------------------
    ----------------------------------------------------------------
*/
struct from_ntup_struct {
  int   status;                   /* NTUP_XX */
  int   tuples_allocated;         /* allocated length of tuple_list */
  int   tuples_found;             /* how much of tuple_list we filled in */
  /* Note: tuple_list has a defined value only if tuples_allocated > 0. */
  struct tuple_struct *tuples;
};

/*  ---------- Constants
*/

/* Note: some code explicitly assumes that WHITE==0, BLACK==1, COLORS==2.
   Terminology: a "color" is WHITE or BLACK.
   A "pseudocolor" is WHITE, BLACK, PSEUDOCOLOR_NOMATCH, or PSEUDOCOLOR_MATCH.
*/
#define WHITE                   0
#define BLACK                   1
#define COLORS                  2

#define PSEUDOCOLOR_NOMATCH     2       /* pixel that matches nothing */
#define PSEUDOCOLOR_MATCH       3       /* pixel that matches everything */
#define PSEUDOCOLOR_MIN         0
#define PSEUDOCOLOR_MAX         3

#define PQ_QTHENP               0
#define PQ_SPECIFIC             1
#define PQ_QTHENP_NOMAXP        2       /* do not use maximize p feature */

#define TUPLE_BLACKONLY         0
#define TUPLE_WHITEONLY         1
#define TUPLE_BLACKWHITE        2

/*  When to give periodic status updates.
*/
#define UPDATE_STRATEGIC        (-2)  /* update at strategic points */
#define UPDATE_NONE             (-1)  /* no updates */
#define UPDATE_OFTEN            0     /* every iteration */
/* update interval is in range UPDATE_MIN.. MAX_INT */
#define UPDATE_MIN              (-2)

/*  Masks for ioflags field of to_ntup_struct.
*/
/* Display input upon entry to generator? */
#define IOF_REPORT_INPUT              (0x1)
/* Display results upon exit from generator? */
#define IOF_REPORT_RESULTS            (0x2)
/* Display generator statistics upon exit from generator? */
#define IOF_NO_TIME_DEPENDENT_OUTPUT  (0x4)
/* Other bit fields are reserved for future use. */

/* What the backtrackers do (subject to back_max_nodes constraint).
*/
#define BACK_ACTION_COUNT       0   /* go through the entire search */
#define BACK_ACTION_ONE         1   /* quit after one solution */

/* What the backtrackers print (specify exactly one of these).
*/
#define BACK_PRINT_NONE         0   /* print no solutions */
#define BACK_PRINT_FIRST        1   /* print first solution found */
#define BACK_PRINT_ALL          2   /* print every solution found */
#define BACK_PRINT_CANON_ONLY   3   /* print every canonical solutions found */
/* note: for back2, BACK_PRINT_CANON_ONLY is equivalent to BACK_PRINT_ALL */
/* note: if BACK_ACTION_COUNT is specified, a summary will be printed
   at the end of the search, regardless of the value of back_print.
*/
#define BACK_PRINT_MAX          3   /* BACK_PRINT is in range 0 .. 3 */

/*  Status codes from ntup_find_xx() functions.
    Must be in range 0 .. NTUP_MAX_RET.
    For meanings, see ntup_print_error().
*/
#define NTUP_OK                                     0
#define NTUP_NO_POS_GIVEN                           1
#define NTUP_NO_NEG_GIVEN                           2
#define NTUP_INVALID_N                              3
#define NTUP_INVALID_TABU_ITERATIONS                4
#define NTUP_INVALID_PQKIND                         5
#define NTUP_INVALID_PORQ                           6
#define NTUP_PI_TOO_SMALL                           7
#define NTUP_INVALID_BACK_MAX_NODES                 8
#define NTUP_INVALID_UPDATE_INTERVAL                9
#define NTUP_INVALID_TUPLE_KIND                     10
#define NTUP_INVALID_BACK2_TIMESLICE_NODES          11
#define NTUP_INVALID_BACK_ACTION                    12
#define NTUP_INVALID_BACK_PRINT                     13
#define NTUP_INVALID_BACK2_N                        14
#define NTUP_INVALID_BACK2_RESTRICT                 15
#define NTUP_INVALID_WIDE1_WIDTH                    16
#define NTUP_INVALID_WIDE1_N                        17
#define NTUP_INVALID_PIXEL_VALUE                    18
#define NTUP_INVALID_POS_BG                         19
#define NTUP_INVALID_NEG_BG                         20
#define NTUP_INVALID_PI_DIMENSIONS                  21
/* NTUP_MAX_RET should be the maximum NTUP_XX value that appears above */
#define NTUP_MAX_RET                                21



#if     defined(__STDC__) || defined(__GNUC__)
#define P(xx) xx
#else
#define P(xx) ()
#define const
#define volatile
#endif

void ntup_print_error P(( FILE *, const struct from_ntup_struct * ));
void ntup_from_free P(( struct from_ntup_struct * ));
void ntup_print P(( FILE *, const struct pix_struct *, const int, const int ));

struct character_struct * ntup_char_trim P(( struct character_struct * const,
                                  const int, void *(*) P((const int)) ));
void ntup_print_char_pixels
  P((FILE *const,const struct character_struct * const));

/* each generator's interface
*/
void ntup_find_back1 P(( const struct to_ntup_struct * const,
  struct from_ntup_struct * const, const int, const int, const double));
void ntup_find_back2 P(( const struct to_ntup_struct * const,
  struct from_ntup_struct * const,
  const int, const int, const double, const double, const int ));
void ntup_find_tabu P(( const struct to_ntup_struct * const,
  struct from_ntup_struct * const, const int ));
void ntup_find_wide1 P(( const struct to_ntup_struct *const,
  struct from_ntup_struct *const, const int));
void ntup_trans_make_rectangular P((struct character_struct *const,const int,
  const int,const int,const int,void  *(*allocfunc) P((const int)) ));

#endif /* _181193_002646 */
