/*
  -------------------- file bitarray.c
  Bit array (multi-word) manipulation routines.
  Created 30 May 93 by Andrew Shapira, Rensselaer Polytechnic Institute.

	Copyright (C) 1993 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/

/*
  In an n-bit array:
    Bit positions are numbered 0 .. n-1.
    Words are numbered 0 .. len-1, where len = BIT_WORDS(n).

  ////////////////////////////////////////////////////////////////////////////
  ///// USERS MUST CALL bit_global_init() before calling these routines.
  ////////////////////////////////////////////////////////////////////////////

  ////////////////////////////////////////////////////////////////////////////
  ///// NOTE: the functions in this library may assume that the bits
  ///// in the unused portion of a bit string are 0.  (For example, on a
  ///// 32-bit machine, a 40-bit string will have 64-40=24 unused bits.)
  ///// These bits must be zeroed prior to using these functions.  A
  ///// typical way of doing things is to allocate a string using bit_alloc(),
  ///// and then initialize it using bit_zero() or bit_ones().  There is some
  ///// more information about this in bit_alloc().
  ////////////////////////////////////////////////////////////////////////////

*/

#define SRCFILE_LIBPROG_BITARRAY

#include "our_common.h"
#include "bitword.h"
#include "bitarray.h"

/*  Initialize tables used by bit manipulation functions.  Users can call
    this function multiple times with the expectation that this function
    and its descendants will do almost nothing, if this function has
    been called previously.
*/
void
bit_global_init()
{
  bitword_global_init();
}

/*  Allocate space and set up pointers for a bit array.  We return a
    pointer m, such that for each i, i = 0 .. rows-1, m[i] is a pointer
    to a bit array of n bits.
*/
BITWORD **
mat_bit_alloc(
const int rows,
const int n,
void  *(*allocfunc) P((const int))
)
{
  const int words = BIT_WORDS(n);
  const int words_per_row = words * sizeof(BITWORD);
  int i;
  char *b1,*b2;
  BITWORD **m;

  b1 = (*allocfunc) ( rows * (sizeof(BITWORD *)+words_per_row) );
  m = (BITWORD **) b1;
  b2 = & b1 [rows*sizeof(BITWORD *)];
  for (i = 0; i < rows; i++)
    m[i] = (BITWORD *) (& b2[i*words_per_row]);
  return(m);
}

/*  Free space allocated by mat_bit_alloc().  */
void
mat_bit_free(
BITWORD **m,
void    (*freefunc) P((void *))
)
{
  (*freefunc) (m);
}


/*  Print the words associated with a bit array of length n.
    Words are printed in hex.
    Does not print trailing newline.

    MACHINE DEPENDENT
*/
void
bit_print_hex(
FILE      *f,
BITWORD   *a,
const int n
)
{
  const int words = BIT_WORDS(n);
  int i;

  for (i = 0; i < words; i++) {
    const BITWORD w = a[i];
    if (i > 0)  fprintf(f, " ");
    fprintf(f, "%04x %04x", (w & 0xffff0000) >> 16, (w & 0x0000ffff) );
  }
}

/*  Print the words associated with a bit array of length n.
    Words are printed in binary.
    Does not print trailing newline.
*/
void
bit_print_binary(
FILE            *f,
const BITWORD   *a,
const int       n
)
{
  const int words = BIT_WORDS(n);
  int i,j;

  for (i = 0; i < words; i++) {
    const BITWORD w = a[i];
    if (i > 0)
      fprintf(f," ");
    for (j = 0; j < sizeof(BITWORD) * 8; j++) {
      if ((j>0) && (j % 8 == 0))  fprintf(f, " ");
      if (BITWORD_IS_BIT_SET(w, (sizeof(BITWORD) * 8 - 1) - j))
        fprintf(f, "1");
      else
        fprintf(f, "0");
    }
  }
}

/*  Make a vector from a bit array.  The bit array has a 1 in position p
    if element p is to appear in the list, for each p,
    0 \leq p \leq BIT_WORDS(n)*sizeof(BITWORD).

    The elements are returned in increasing order.
    NOTE: one way to sort a vector is to make a bit array from
    the vector using bit_make_from_vec(), and then use bit_make_vec()
    to get the elements back in order.  This will be very efficient
    for some kinds of vectors.  See also sort.c/intsort().

    list_ret should be large enough to hold the number of elements we find.

    Is compatible with bit_make_from_vec().

    COULDDO: speed up by inserting the code from FOR_ONE_BITS here
    directly, then processing one word at a time without the various
    recalculations that go on.
*/
void
bit_make_vec(
const BITWORD *a,
const int n,
int *list_ret,        /* we place the list here */
int *listlen_ret      /* number of elements we put in list */
)
{
  int len;            /* our version of list length */
  int word, pos;

  len = 0;
  FOR_ONE_BITS(a,n,pos,word) /* { */
    list_ret[len++] = pos;
  }

  *listlen_ret = len;
}

/*  Build a bit array from a matl.
    For each i, 0 <= i < range:
        bit i is 1  if  the element i+min appears somewhere in m.
                 0  otherwise
*/
void
bit_make_from_matl(
BITWORD *b_ret,             /* allocated length should be nvals or more */
int **m,                    /* const */
const int rows,
const int *row_lengths,
const int nvals             /* values of m must be in range 0 .. nvals-1 */
)
{
  int row,col;

  bit_zero(b_ret, nvals);
  for (row = 0; row < rows; row++) {
    const int len = row_lengths[row];
    const int *thisrow = m[row];

    for (col = 0; col < len; col++) {
      const int element = thisrow[col];
      BIT_SET(b_ret, element);
    }
  }
}

/*  The rank of a bit is the number of 1's that are in smaller positions.

    This function finds the position of the rank r bit of b.
    Returns:   0..n-1     position of the rank r bit, if it exists
              -1          if fewer than r+1 bits are set in b
    
    COULDDO: use bit_count to count multiple bits at once
*/
int
bit_find_by_rank(
const BITWORD *b,
const int n,
const int r
)
{
  int rank;
  int pos,word;

  rank = 0;
  FOR_ONE_BITS(b,n,pos,word) /* { */
    if (rank == r)  return(pos);
    rank++;
  }
  return(-1);
}

/*
  Compares two bit strings that are of equal length.
  Returns:
      -1      a < b
      0       a = b
      1       a > b
    Is compatible with bit_find_smallest() and bit_find_largest().
    Can be used as comparison function to qsort().
*/
int
bit_compare(
const BITWORD *a,
const BITWORD *b,
const int n
)
{
  const int words = BIT_WORDS(n);
  int word;

  for (word = 0; word < words; word++) {
    const int compare = bitword_compare(a[word],b[word]);
    if (compare)  return(compare);
  }
  return(0);
}

