/*
  -------------------- file memalloc.c

	Copyright (C) 1992 Andrew Shapira.

	This file is part of the Ntuplib library.

	Ntuplib is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation; either version 2 of the License, or
	(at your option) any later version.

	Ntuplib is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with the Ntuplib Library; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <stdio.h>
#include <malloc.h>


#define SRCFILE_LIBPROG_MEMALLOC

#include "our_common.h"
#include "memalloc.h"
#include "util.h"



/* --------------------------------------------------------
  Routines for the manipulation of a particular stack, called the
  "buffer stack".  Use this for fast allocation of memory.

  TODO:
    Make a way to reclaim all the space allocated by these functions and
    restore everything to its original state.
*/

#ifdef  BSDEBUG
#define MIN_POOL_SIZE 8
#define INITIAL_POOLS 1
#define INITIAL_USED  1
#else   /*BSDEBUG*/
#define MIN_POOL_SIZE       1024  /* each pool is at least this big */
#define INITIAL_POOLS       16    /* must be >= 1 */
#define INITIAL_USED        16    /* must be >= 1 */
#endif  /*BSDEBUG*/

int _memalloc_depth = -1;       /* stack depth */
static int  totallocated = 0;   /* total allocated bytes */
int *_memalloc_used;            /* how many bytes used at each stack depth */
int _memalloc_used_alloc = 0;   /* allocation of _memalloc_used[] */
int _memalloc_used_at_deepest = 0;  /* invar: memalloc_used[_memalloc_depth] */
static int  highwater;          /* most we have allocated, over all time */

struct _memalloc_pool_struct *_memalloc_pool;
static int  _memalloc_pool_alloc = 0;
int _memalloc_pool_depth = -1;
/* invariants for pool structure at deepest level */
int _memalloc_pool_used_at_deepest = 0;
int _memalloc_pool_allocated_at_deepest = 0;
unsigned char *_memalloc_pool_buffer_at_deepest = NULL;

/* where all allocations of zero length point to */
unsigned char _memalloc_zerolenchar;


#ifdef  BSDEBUG
void
bsdebug(const char *str)
{
  int   i;

  if (str)
    printf("\n--------- begin bsdebug dump:    %s\n", str);
  else
    printf("\n--------- begin bsdebug dump\n");
  printf( "depth=%d used_alloc=%d totalloc=%d highwat=%d",
    _memalloc_depth, _memalloc_used_alloc,totallocated,highwater);

  printf(" pool_alloc=%d pool_depth=%d\n",
    _memalloc_pool_alloc,_memalloc_pool_depth);
  for (i = -1; i <= _memalloc_pool_depth-1; i++) {
    printf("pool[%2d]      :  buffer=@%8d  used=%7d  allocated=%7d\n",
      i, (int)_memalloc_pool[i].buffer, _memalloc_pool[i].used,
      _memalloc_pool[i].allocated);
  }

  printf("pool[deepest] :  buffer=@%8d  used=%7d  allocated=%7d\n",
    (int) _memalloc_pool_buffer_at_deepest, _memalloc_pool_used_at_deepest,
    _memalloc_pool_allocated_at_deepest);
  for (i = -1; i <= _memalloc_depth-1; i++)
    printf("used[%2d]      = %6d\n", i, _memalloc_used[i]);
  printf("used[deepest] = %6d\n", _memalloc_used_at_deepest);

  printf("--------- end bsdebug dump\n\n");
}
#endif  /*BSDEBUG*/


void
_memalloc_do_used_alloc()
{
  BSD(printf("do_used_alloc()\n");)
  if (! _memalloc_used_alloc) {                 /* full initialization */
    _memalloc_used_alloc = INITIAL_USED;
    _memalloc_used =
              xmalloc((_memalloc_used_alloc+1)*sizeof(_memalloc_used[0]));
    _memalloc_used++;
    _memalloc_used[-1] = 0;
  }
  else {
    _memalloc_used_alloc *= 2;
    _memalloc_used = xrealloc(_memalloc_used-1,
              (_memalloc_used_alloc+1)*sizeof(_memalloc_used[0]));
    _memalloc_used++;
  }
  /*  _memalloc_used now accessed in range [ -1 .. _memalloc_used_alloc-1] */
}

static void
do_pool_alloc()
{
  BSD(printf("do_pool_alloc()\n");)
  if (! _memalloc_pool_alloc) {
    _memalloc_pool_alloc = INITIAL_POOLS;
    _memalloc_pool = xmalloc(
                        (_memalloc_pool_alloc+1)*sizeof(_memalloc_pool[0]));
    _memalloc_pool++;
    _memalloc_pool[-1].buffer = NULL;
    _memalloc_pool[-1].allocated = 0;
    _memalloc_pool[-1].used = 0;
  }
  else {
    _memalloc_pool_alloc *= 2;
    _memalloc_pool = xrealloc((_memalloc_pool-1),
                        (_memalloc_pool_alloc+1)*sizeof(_memalloc_pool[0]));
    _memalloc_pool++;
  }
  /*  _memalloc_pool now accessed in range [ -1 .. _memalloc_pool_alloc-1] */
}

/* This is the only place where memory is actually allocated for the user. */
void
_memalloc_nextpool(
const int get
)
{
  int                 nalloc;

  BSD(printf("nextpool()\n");)
  if (_memalloc_pool_depth >= 0) {
    struct _memalloc_pool_struct  *p;
    p = & _memalloc_pool[_memalloc_pool_depth];
    p->used = _memalloc_pool_used_at_deepest;
    p->allocated = _memalloc_pool_allocated_at_deepest;
    p->buffer = _memalloc_pool_buffer_at_deepest;
  }

  _memalloc_pool_depth++;
  if (_memalloc_pool_depth >= _memalloc_pool_alloc)
    do_pool_alloc();

  /*TODO: there is probably a better system */
  /* One problem is that if you have, say:  pool 0  75000 / 80000 used
      and you want 10K for the next bsbegin context, it will allocate
      80000, which is fine, but then if you hit a bsend, a bsbegin, and
      want another 10K, what you will get is a deallocation and allocation
      of 80K again.  This could be avoided.  One way is to defer the
      actual deallocation of the 80K until the next memory allocation
      request.
  */
  if (_memalloc_pool_depth > 0) {
    nalloc = MAX(totallocated, 2*get);
  }
  else {
    nalloc = MAX(highwater, 2*get);
  }
  nalloc = MAX(nalloc, MIN_POOL_SIZE);
#ifdef  OLD_ALLOC_XXX
  nalloc = MAX(highwater, 2*get);
  nalloc = MAX(nalloc, MIN_POOL_SIZE);
#endif  /*OLD_ALLOC_XXX */

  _memalloc_pool_used_at_deepest = 0;
  _memalloc_pool_allocated_at_deepest = nalloc;
  _memalloc_pool_buffer_at_deepest = xmalloc(nalloc);

  totallocated += nalloc;
  highwater = MAX(highwater, totallocated);
}

/* This is the only place where memory allocated for the user is reclaimed. */
void
_memalloc_reclaim()
{
  BSD(printf("reclaim()\n");)
  while ((_memalloc_pool_depth >= 0) &&
              (_memalloc_used_at_deepest >= _memalloc_pool_used_at_deepest)) {
    if ((_memalloc_used_at_deepest == _memalloc_pool_used_at_deepest) &&
                    (_memalloc_pool_allocated_at_deepest == highwater))
      break;
    _memalloc_used_at_deepest -= _memalloc_pool_used_at_deepest;
    totallocated -= _memalloc_pool_allocated_at_deepest;
    BSD(printf("--- bsreclaim: freeing pool %d (@%d)\n",
      _memalloc_pool_depth, (int) (_memalloc_pool_buffer_at_deepest));)
    xfree(_memalloc_pool_buffer_at_deepest);
    {
      struct _memalloc_pool_struct *p;

      _memalloc_pool_depth--;
      p = & _memalloc_pool[_memalloc_pool_depth];
      _memalloc_pool_used_at_deepest = p->used;
      _memalloc_pool_allocated_at_deepest = p->allocated;
      _memalloc_pool_buffer_at_deepest = p->buffer;
    }
  }
  if (_memalloc_pool_depth >= 0)
    _memalloc_pool_used_at_deepest -= _memalloc_used_at_deepest;
}

